/*
 * edu.calstatela.curvebank.goodstein.Applet
 * Goodstein Sequence Applet
 * 
 * Created December 16, 2003
 * 
 * Send any comments or bug fixes to the author at matthew.nelson@acm.org.
 * 
 * Copyright (c) 2004 Matthew Nelson. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *   1. Redistributions of source code must retain the above copyright notice,
 *      this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *   3. The name of the author may not be used to endorse or promote products
 *      derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 
package edu.calstatela.curvebank.goodstein;

import edu.calstatela.curvebank.goodstein.GoodsteinSequence.Element;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.table.TableModel;
import javax.swing.table.TableColumnModel;
import java.math.BigInteger;

/**
 * An applet or stand-alone application that demonstrates the Goodstein
 * Sequence. The interface allows the user to select two values and begin
 * the computation of the Goodstein Sequence. The bulk of the work is performed
 * in {@link BackgroundThread}, which sends messages to this class when new
 * elements are available. Finally, a {@link GoodsteinTableModel} stores the
 * computed values.
 * 
 * @author Matthew Nelson
 */
public class Applet extends JApplet {

	static {
		try {
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
		} catch (Exception unused) {}
	}
		
	private JSpinner numberComp;
	private JSpinner radixComp;
	private JTable outputTable;
	private JScrollPane outputScrollPane;	
	private JButton startButton;
	private JButton stopButton;

	private BackgroundThread backgroundThread;
	
	/**
	 * Creates a JFrame that contains an instance of this class and invokes
	 * the appropriate <code>java.applet.Applet</code> methods.
	 */
	public static void main(String[] args)
	{
		final Applet applet = new Applet();
		
		// Create a JFrame that automatically calls the proper applet metods
		JFrame frame = new JFrame() {
			protected void processWindowEvent(WindowEvent e) {
				if (e.getID() == WindowEvent.WINDOW_OPENED) {
					applet.start();
				} else if (e.getID() == WindowEvent.WINDOW_CLOSING) {
					applet.stop();
					applet.destroy();
					dispose();
				} else if (e.getID() == WindowEvent.WINDOW_CLOSED) {
					System.exit(0); // End process
				} else {
					super.processWindowEvent(e);
				}
			}
		};
	
		// Initialize
		frame.setTitle("Goodstein Sequence Applet");
		applet.setSize(475, 475); // For properly initializing column widths
		applet.init();
		frame.getContentPane().add(applet);
		frame.pack();
		frame.setLocationRelativeTo(null); // Center
	
		frame.setVisible(true);
	}
	
	/*
	 * @see java.applet.Applet#init()
	 */
	public void init() {
		
		numberComp = new JSpinner(new SpinnerNumberModel(3, 0, 999999, 1));
		radixComp = new JSpinner(new SpinnerNumberModel(2, 2, 999999, 1));
				
		startButton = new JButton("Start");
		startButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				doStart();
			}
		});

		stopButton = new JButton("Stop");
		stopButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				doStop();
			}
		});

		Box buttonBox = new Box(BoxLayout.X_AXIS);
		buttonBox.add(new JLabel("Number:"));
		buttonBox.add(Box.createHorizontalStrut(5));
		buttonBox.add(numberComp);
		buttonBox.add(Box.createHorizontalGlue());
		buttonBox.add(Box.createHorizontalStrut(10));
		buttonBox.add(new JLabel("Base:"));
		buttonBox.add(Box.createHorizontalStrut(5));
		buttonBox.add(radixComp);
		buttonBox.add(Box.createHorizontalGlue());
		buttonBox.add(Box.createHorizontalStrut(10));
		buttonBox.add(startButton);
		buttonBox.add(stopButton);
		buttonBox.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));

		// Create the JTable with a slightly modified DefaultTableModel
		TableModel outputModel = new GoodsteinTableModel();
		outputTable = new JTable(outputModel);
		outputTable.setAutoResizeMode(JTable.AUTO_RESIZE_NEXT_COLUMN);
		outputTable.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		TableColumnModel columnModel = outputTable.getColumnModel();
		columnModel.getColumn(0).setPreferredWidth(50);
		columnModel.getColumn(1).setPreferredWidth(50);
		columnModel.getColumn(2).setPreferredWidth(getWidth()/2 - 50);
		columnModel.getColumn(3).setPreferredWidth(getWidth()/2 - 50);
		
		// Create the scroll pane that views the JTable
		outputScrollPane = new JScrollPane(outputTable);
		outputScrollPane.getVerticalScrollBar().getModel();
		
		getContentPane().add(buttonBox, BorderLayout.NORTH);
		getContentPane().add(outputScrollPane, BorderLayout.CENTER);

	}
	
	/**
	 * When the user views a page with this applet, the controls are (re)set
	 * as if a background thread had just been stopped. Additionally, any
	 * rows in the JTable are cleared.
	 *  
	 * @see java.applet.Applet#start()
	 */
	public void start() {
		
		// Invoke stop action
		doStop();
		
		// Clear JTable
		GoodsteinTableModel outputModel
				= (GoodsteinTableModel) outputTable.getModel();
		outputModel.clear();
		
	}
	
	/**
	 * Should the user exit a page, this method automatically stops the
	 * background thread.
	 * 
	 * @see java.applet.Applet#stop()
	 */
	public void stop() {
		
		// Invoke stop action
		doStop();
		
	}
	
	private void doStart() {
		
		// Clear JTable's model of all rows
		GoodsteinTableModel outputModel
				= (GoodsteinTableModel) outputTable.getModel();
		outputModel.clear();
		outputModel.setInitialBase(
				((Integer) radixComp.getValue()).intValue());
		
		// Disable controls and enable the Stop button
		startButton.setEnabled(false);
		numberComp.setEnabled(false);
		radixComp.setEnabled(false);
		stopButton.setEnabled(true);
		
		// Start background thread to do processing,
		// passing it the values from the spinners
		backgroundThread = new BackgroundThread(
				new BigInteger(numberComp.getValue().toString()),
				new BigInteger(radixComp.getValue().toString()), this,
				10000); // *** Limit to 10,000 steps ***
		backgroundThread.setPriority(Thread.MIN_PRIORITY);
		backgroundThread.start();

	}
	
	private void doStop() {
		
		// Stop current background thread
		if (backgroundThread != null) {
			backgroundThread.interrupt();
			backgroundThread = null;
		}
		
		// Disable Start button and enable controls
		stopButton.setEnabled(false);
		startButton.setEnabled(true);
		numberComp.setEnabled(true);
		radixComp.setEnabled(true);

	}
	
	void doElementAdd(Element e, Thread t) {
		if (t == backgroundThread) {
			// Add element to table model
			GoodsteinTableModel outputModel
					= (GoodsteinTableModel) outputTable.getModel();
			outputModel.addRow(e.getNumber(), e.toString());
		}
	}
	
	void doSequenceComplete(Thread t) {
		if (t == backgroundThread) {
			doStop();
		}
	}

}
