/*
 * edu.calstatela.curvebank.goodstein.BackgroundThread
 * Goodstein Sequence Applet
 * 
 * Created December 17, 2003
 * 
 * Send any comments or bug fixes to the author at matthew.nelson@acm.org.
 * 
 * Copyright (c) 2004 Matthew Nelson. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *   1. Redistributions of source code must retain the above copyright notice,
 *      this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *   3. The name of the author may not be used to endorse or promote products
 *      derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package edu.calstatela.curvebank.goodstein;

import edu.calstatela.curvebank.goodstein.GoodsteinSequence.Element;
import javax.swing.SwingUtilities;
import java.math.BigInteger;

/**
 * Serves {@link Applet} by computing the values of the Goodstein Sequence and
 * sending messages on the AWT Event Dispatch Thread as the elements become
 * available.
 * 
 * <p>The security manager threw some exceptions on some operating systems when
 * the <code>interrupt()</code> method was executed, so this class implements
 * its own interrupt state that circumvents the security manager. This means
 * that any thread may invoke the <code>interrupt()</code> method of an
 * executing <code>BackgroundThread</code>, and <code>interrupt()</code> behaves
 * differently internally.
 * 
 * @author Matthew Nelson
 */
class BackgroundThread extends Thread {
	
	/**
	 * Reference to parent <code>Applet</code> object. Messages concerning the
	 * generation of Goodstein Sequence elemetns are sent to this object.
	 */
	private Applet parent;
	
	/**
	 * The <code>GoodsteinSequence</code> that provides the elements of the
	 * Goodstein Sequence.
	 */
	private GoodsteinSequence sequence;
	
	private final int limit;
	
	/**
	 * Creates a new background thread that returns the numbers of the
	 * Goodstein Sequence with first element <code>n</code> in base
	 * <code>b</code> notation.
	 * 
	 * @param n the first element of the Goodstein Sequence
	 * @param b the base of the first element
	 * @param a the parent applet to receive messages
	 * @param limit the maximum number of elements returned, or 0 to indicate
	 *          no limit
	 */
	BackgroundThread(BigInteger n, BigInteger b, Applet a, int limit) {
		sequence = new GoodsteinSequence(n, b);
		parent = a;
		this.limit = limit;
	}
	
	/*
	 * @see java.lang.Runnable#run()
	 */
	public void run() {
		int counter = 0;
		while (true) {

			// If the Goodstein sequence has more elements... 
			if (sequence.hasMoreElements() && !isInterrupted()) {
				try {
				
					// ...find the next one
					final Element element = (Element) sequence.nextElement();
					
					// .. and notify the GUI on the AWT Event thread
					if (!isInterrupted()) {
						SwingUtilities.invokeAndWait(new Runnable() {
							public void run() {
								parent.doElementAdd(
										element, BackgroundThread.this);
							}
						});
					}

				} catch (Throwable e) {
					e.printStackTrace();
					// Some error or exception occurred
					// Interrupt self
					interrupt();
				}
			}
			
			// Otherwise, the Goodstein sequence is complete
			else {
				// Interrupt self
				interrupt();
			}

			++counter;
			
			// If this thread has been interrupted...
			if (isInterrupted() || (limit > 0 && counter >= limit)) {
				// ... notify the GUI on the AWT Event thread
				try {
					SwingUtilities.invokeAndWait(new Runnable() {
						public void run() {
							parent.doSequenceComplete(BackgroundThread.this);
						}
					});
				} catch (Exception unused) {}
		
				// ... and automatically end this thread
				return;
			}
		}
	}
	
	private boolean isInterruptedFlag = false;
	
	/*
	 * @see java.lang.Thread#isInterrupted()
	 */
	public synchronized boolean isInterrupted() {
		return isInterruptedFlag;
	}
	
	/*
	 * @see java.lang.Thread#interrupt()
	 */
	public synchronized void interrupt() {
		isInterruptedFlag = true;
	}
		
}
