/*
 * edu.calstatela.curvebank.goodstein.GoodsteinSequence
 * Goodstein Sequence Applet
 * 
 * Created December 17, 2003
 * 
 * Send any comments or bug fixes to the author at matthew.nelson@acm.org.
 * 
 * Copyright (c) 2004 Matthew Nelson. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *   1. Redistributions of source code must retain the above copyright notice,
 *      this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *   3. The name of the author may not be used to endorse or promote products
 *      derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package edu.calstatela.curvebank.goodstein;

import java.util.Enumeration;
import java.util.NoSuchElementException;
import java.math.BigInteger;

/**
 * An enumeration that returns the elements of the Goodstein Sequence in order.
 * The sequence is backed by a instance of {@link HereditaryBaseNumber}.
 * Invokations of {@link #nextElement()} will take some time for larger numbers.
 * 
 * @author Matthew Nelson
 */
public class GoodsteinSequence implements Enumeration {
	
	private HereditaryBaseNumber term; 
	private BigInteger index = BigInteger.ZERO;
	private boolean firstTermReturned = false;
	
	/**
	 * Constructs a new sequence with the given first element, using superbase
	 * <code>base</code> notation initially.
	 * 
	 * @param firstElement the first number of the sequence
	 * @param base the base of the first element
	 */
	public GoodsteinSequence(BigInteger firstElement, BigInteger base) {
		term = new HereditaryBaseNumber(firstElement, base);
	}
	
	/**
	 * Constructs a new sequence with the given first element, using superbase
	 * 2 notation initially.
	 * 
	 * @param firstElement the first number of the sequence 
	 */
	public GoodsteinSequence(BigInteger firstElement) {
		term = new HereditaryBaseNumber(firstElement, new BigInteger("2"));
	}
	
	/**
	 * Determines whether this sequence has more elements.
	 * 
	 * @return true if the next invokation of {@link #nextElement()} will
	 *         not throw a <code>NoSuchElementException</code>
	 */
	public boolean hasMoreElements() {
		return term.compareTo(BigInteger.ZERO) > 0
				|| !firstTermReturned;
	}
	
	/**
	 * Returns the next element of this Goodstein Sequence. This computation
	 * may take some time for larger numbers. If no more elements are available,
	 * then this method throws an exception.
	 * 
	 * @return an {@link Element} representing the next element
	 * @throws NoSuchElementException if no more elements are available
	 */
	public Object nextElement() {
		if (!hasMoreElements()) {
			throw new NoSuchElementException();
		}
		if (firstTermReturned) {
			index = index.add(BigInteger.ONE);
			term.incrementRadix();
			term.decrement();
		} else {
			firstTermReturned = true;
		}

		// The constructor of inner class reads the state of this sequence
		return new Element();
	}
	
	/**
	 * Contains all the information related to a single element of the Goodstein
	 * Sequence. All data is cached as {@link #nextElement()} executes, so
	 * that this class's methods will always execute quickly.
	 * 
	 * @author Matthew Nelson
	 */
	public class Element {
		private BigInteger index;
		private BigInteger base;
		private BigInteger number;
		private String string;
		
		private Element() {
			index = GoodsteinSequence.this.index;
			base = GoodsteinSequence.this.term.getRadix();
			number = GoodsteinSequence.this.term.getValue();
			string = GoodsteinSequence.this.term.toString();
		}
		
		/**
		 * Returns the index of this element. The first element always has
		 * index zero.
		 * 
		 * @return the index of this element
		 */
		public BigInteger getIndex() {
			return index;
		}
		
		/**
		 * Returns the base used in the superbase notation of this element.
		 * 
		 * @return the base used for this element
		 */
		public BigInteger getBase() {
			return base;
		}
		
		/**
		 * Returns this element's value. This number is cached, so execution
		 * of this method is always fast.
		 * 
		 * @return this element's value
		 */
		public BigInteger getNumber() {
			return number;
		}
		
		/**
		 * Returns a string representation of this element in superbase
		 * notation.
		 * 
		 * @return a string representation of the superase number
		 */
		public String toString() {
			return string;
		}
	}
	
}