/*
 * edu.calstatela.curvebank.goodstein.HereditaryBaseNumber
 * Goodstein Sequence Applet
 * 
 * Created December 17, 2003
 * 
 * Send any comments or bug fixes to the author at matthew.nelson@acm.org.
 * 
 * Copyright (c) 2004 Matthew Nelson. All Rights Reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 *   1. Redistributions of source code must retain the above copyright notice,
 *      this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *   3. The name of the author may not be used to endorse or promote products
 *      derived from this software without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
 * EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 
package edu.calstatela.curvebank.goodstein;

import java.util.Stack;
import java.math.BigInteger;

/**
 * Represents some natural number in superbase notation and provides for
 * decrement operations on the number and increment operations on the base.
 * Those two operations are essential to producing the numbers of the
 * Goodstein sequence.
 * 
 * <p>This implementation was designed with a high level of efficiency in mind.
 * The most computationally intensive tasks involve computing the value
 * represented by instances of this class. A caching mechanism remembers the
 * value whenever possible to greatly improve efficiency, but invokations of
 * {@link #incrementRadix()} will clear the cache. Once this happens,
 * the next call to {@link #getValue()} or possibly a call to
 * {@link #equals(Object)} or {@link #compareTo(Object}} will force a
 * lengthy recomputation of the value.
 * 
 * @author Matthew Nelson
 */
public class HereditaryBaseNumber implements Comparable {
	
	/*
	 * The notation is represented by a tree of HereditaryBaseNumber and Term
	 * objects. The root object is a HereditaryBaseNumber that contains a list
	 * of all Term objects in the next level of the tree. Each Term object
	 * maintains a reference to a nested HereditaryBaseNumber, which
	 * represents the value of the exponent. The leaf nodes of this tree are all
	 * HereditaryBaseNumber objects that represent 0 with an empty list of
	 * terms. 
	 */
	
	/**
	 * The lowest-order term in a linked list of terms. Each term maintains
	 * a reference to the next term, sorted in ascending order. Or, a term may
	 * reference to <code>null</code> if it represents the term of greatest
	 * order. If this reference is <code>null</code>, then the value
	 * represented by this <code>HereditaryBaseNumber</code> object is zero.
	 * Any terms with a zero coefficient are <b>always</b> removed from this
	 * list. Furthermore, the coefficients of these terms are never greater than
	 * or equal to the radix.</b>
	 */
	private Term lowestOrderTerm;

	/**
	 * A reference to a single object that maintains the radix value for an
	 * entire tree of <code>HereditaryBaseNumber</code> and {@link Term}
	 * objects.
	 */
	private SharedValue radix = null;
	
	/**
	 * Constructs a new number in hereditary base notation from the given
	 * number with the given radix. See {@link #setNumber(BigInteger,
	 * BigInteger)} for information on arguments and the exceptions
	 * possibly thrown by this constructor.
	 * 
	 * @see #setNumber(BigInteger, BigInteger)
	 */
	public HereditaryBaseNumber(BigInteger number, BigInteger radix) {
		setNumber(number, radix);
	}

	/**
	 * A private constructor that allows for optimizations in internal
	 * algorithms by bypassing the unnecessary algorithm in the public
	 * constructor.
	 */
	private HereditaryBaseNumber() {}
	
	/**
	 * Regenerates the superbase expression represented using the specified
	 * value and base.
	 * 
	 * @param number a <i>positive</i> value
	 * @param radix a base <i>greater than or equal to 2</i>
	 * @throws IllegalArgumentException if the arguments do not lie in the
	 *         allowed ranges
	 */
	public void setNumber(BigInteger number, BigInteger radix) {

		// Check arguments
		if (number.compareTo(BigInteger.ZERO) < 0) {
			throw new IllegalArgumentException("number must be positive");
		}
		if (radix.compareTo(new BigInteger("2")) < 0) {
			throw new IllegalArgumentException("radix must be at least 2");
		}

		// Create new global radix object
		SharedValue sharedRadix = new SharedValue(radix);
		
		// Begin recursive generation of tree
		setNumber(number, sharedRadix);
		
	}
	
	/**
	 * A private step in {@link #setNumber(BigInteger, BigInteger)} that
	 * regenerates a subtree with value <code>number</code> while propagating
	 * the same shared <code>radix</code> object.
	 */
	private void setNumber(BigInteger number, SharedValue radix) {
		
		// Copy reference to shared radix object
		this.radix = radix;
		
		// Optimization - Cache the value, which is now known
		this.valueCache = number;
		this.radixExpectedModCount = this.radix.getModCount();
		
		// Clear the lowest order term, since number may be zero
		this.lowestOrderTerm = null;
		
		// Convert number to given base, creating a recursive tree. Terms are
		// created in reverse order -- with the largest order first -- so push
		// onto a stack to reverse later
		Stack terms = new Stack();
		BigInteger position = BigInteger.ZERO;
		while (!number.equals(BigInteger.ZERO)) {
			
			// number /= radix
			// coefficient = number % radix
			BigInteger[] divideResults
					= number.divideAndRemainder(radix.getValue());
			number = divideResults[0];
			BigInteger coefficient = divideResults[1];
			
			// Terms with zero coefficient are  not useful. Don't waste memory
			if (!coefficient.equals(BigInteger.ZERO)) {
				HereditaryBaseNumber exponent = new HereditaryBaseNumber();
				exponent.setNumber(position, radix);
				terms.push(new Term(coefficient, exponent, null));
			}
			
			// ++position
			position = position.add(BigInteger.ONE);

		}
		
		// Reverse the stack and create linked list
		while (!terms.isEmpty()) {
			Term higherOrderTerm = lowestOrderTerm;
			lowestOrderTerm = (Term) terms.pop();
			lowestOrderTerm.next = higherOrderTerm;
		}
		 
	}

	/**
	 * A cache used by {@link #getValue()}. It saves the last computed or known
	 * value represented by this <code>HereditaryBaseNumber</code> object. This
	 * value may be null or old, so local algorithms should <i>always</i>
	 * invoke <code>getValue()</code> to obtain this value. Additionally,
	 * the following methods will update the cache when the value is "known"
	 * to improve efficiency:
	 * 
	 * <ul>
	 * <li>{@link #setNumber(BigInteger, SharedValue)}</li>
	 * <li>{@link #decrement()}</li>
	 * <li>{@link #duplicate()}</li>
	 * </ul>
	 */
	private BigInteger valueCache = null;

	/**
	 * Used internally by {@link #getValue()}. It stores the last "mod count"
	 * of the radix object.
	 */
	private int radixExpectedModCount = 0;

	/**
	 * Returns the value represented by this object. Should the value not be
	 * known, this method leads into a <i>very lengthy</i> computation that
	 * will take some time for large numbers.
	 * 
	 * @return the value represented by this object
	 */	
	public BigInteger getValue() {
		
		// Check for modifications that would force a recalculation
		// (If cache is empty or the radix has changed.)
		if (valueCache == null || radix.getModCount() != radixExpectedModCount) {
			// Update expected mod count
			radixExpectedModCount = radix.getModCount();
			
			// Recalculate value
			Term term = lowestOrderTerm;
			valueCache = BigInteger.ZERO;
			while (term != null) {
				// value += coefficeint * base ^ exponent
				valueCache = valueCache.add(term.coefficient.multiply(
						pow(radix.getValue(), term.exponent.getValue())));
				term = term.next;
			}
		}
		
		return valueCache;
	}
	
	/**
	 * A method used internally by {@link #getValue()} to compute any
	 * <code>BigInteger</code> raised to any <code>BigInteger</code>. <i>This
	 * computation is intensive and will take some time with larger numbers.</i>
	 */
	public static BigInteger pow(BigInteger base, BigInteger exp) {
		if (exp.compareTo(BigInteger.ZERO) < 0) {
			throw new IllegalArgumentException();
		}
		BigInteger result = BigInteger.ONE;
		while (exp.compareTo(BigInteger.ZERO) > 0) {
			result = result.multiply(base);
			exp = exp.subtract(BigInteger.ONE);
		}
		return result;
	}
	
	/**
	 * Returns the radix value used in this superbase notation.
	 * 
	 * @return the radix value
	 */
	public BigInteger getRadix() {
		return radix.getValue();
	}
	
	/**
	 * Increments the base by one. This operation replaces all occurrences of
	 * the number returned by {@link #getRadix()} with the same number plus one. 
	 */
	public void incrementRadix() {	
		radix.increment();
	}

	/**
	 * Subtracts one from the number represented by this object.
	 */	
	public void decrement() {
				
		// Cannot decrement below 0
		if (lowestOrderTerm == null) {
			throw new IllegalArgumentException();
		}
		
		// Optimization - update cached value
		if (valueCache != null) {
			valueCache = valueCache.subtract(BigInteger.ONE);
		}
		
		// If the lowest order term's coefficeint is multiplied by radix^0
		// Then simply subtract one from it. If the new coefficient is zero,
		// then remove the term from the list of terms.
		if (lowestOrderTerm.exponent.equals(BigInteger.ZERO)) {
			lowestOrderTerm.coefficient
					= lowestOrderTerm.coefficient.subtract(BigInteger.ONE);
			if (lowestOrderTerm.coefficient.equals(BigInteger.ZERO)) {
				lowestOrderTerm = lowestOrderTerm.next;
			}
		}
		
		// If the lowest order term has some exponent, then carry values from
		// the left.
		else {
			// Note the lowest order term's power
			HereditaryBaseNumber power = lowestOrderTerm.exponent.duplicate();
			
			// Subtract one from lowest order term's coefficient
			// and remove from list if new coefficient is zero
			lowestOrderTerm.coefficient
					= lowestOrderTerm.coefficient.subtract(BigInteger.ONE);
			if (lowestOrderTerm.coefficient.equals(BigInteger.ZERO)) {
				lowestOrderTerm = lowestOrderTerm.next;
			}
			
			// Create a term for each power less than the original term's power
			BigInteger radixMinusOne = radix.getValue().subtract(BigInteger.ONE);
			do {
				power.decrement();
				lowestOrderTerm = new Term(radixMinusOne, power.duplicate(),
						lowestOrderTerm);
			} while (power.compareTo(BigInteger.ZERO) > 0);
		}

	}
	
	/**
	 * A private method that duplicates the subtree of nodes represented by
	 * this <code>HereditaryBaseNumber</code> object. The resulting subtree
	 * must be used in the same tree, since the nodes still reference
	 * the same global radix value.
	 */
	private HereditaryBaseNumber duplicate() {
		HereditaryBaseNumber clone = new HereditaryBaseNumber();
		
		// Optimization - copy cache
		clone.valueCache = valueCache;
		clone.radixExpectedModCount = radixExpectedModCount;
		
		clone.radix = radix; 
		clone.lowestOrderTerm =
				(lowestOrderTerm == null) ? null : lowestOrderTerm.duplicate();
		
		return clone;
	}
	
	/**
	 * Compares the number represented by this <code>HereditaryBaseNumber</code>
	 * with the value represented by another <code>HereditaryBaseNumber</code>
	 * or with a <code>BigInteger</code>. Execution of this method is very
	 * efficient if either object represents a zero.
	 * 
	 * @throws ClassCastException if <code>obj</code> is not a
	 *           <code>HereditaryBaseNumber</code> or a <code>BigInteger</code>
	 */
	public int compareTo(Object obj) {
		if (obj instanceof HereditaryBaseNumber) {
			// Compare to another HereditaryBaseNumber
			HereditaryBaseNumber n = (HereditaryBaseNumber) obj;
			
			// Special optimizations to accelerate comparisons involving 0
			if (n.lowestOrderTerm == null) {
				return (this.lowestOrderTerm == null) ? 0 : 1;
			} else if (this.lowestOrderTerm == null) {
				return (n.lowestOrderTerm == null) ? 0 : -1;
			}
			
			// Slowest comparison
			return getValue().compareTo(n.getValue());
			
		} else if (obj instanceof BigInteger) {
			// Compare to a BigInteger
			BigInteger n = (BigInteger) obj;
			
			// Special optimizations to accelerate comparisons involving 0
			if (n.equals(BigInteger.ZERO)) {
				return (this.lowestOrderTerm == null) ? 0 : 1;
			} else if (this.lowestOrderTerm == null) {
				return (n.equals(BigInteger.ZERO)) ? 0 : 1;
			}
			
			// Slowest comparison
			return getValue().compareTo(n);
			
		} else {
			throw new ClassCastException();
		}
	}
	
	/**
	 * Compares the number represented by this <code>HereditaryBaseNumber</code>
	 * with the value represented by another <code>HereditaryBaseNumber</code>
	 * or with a <code>BigInteger</code>. Execution of this method is very
	 * efficient if either object represents a zero.
	 * 
	 * @throws ClassCastException if <code>obj</code> is not a
	 *           <code>HereditaryBaseNumber</code> or a <code>BigInteger</code>
	 */
	public boolean equals(Object obj) {
		if (obj instanceof HereditaryBaseNumber) {
			// Compare to another HereditaryBaseNumber
			HereditaryBaseNumber n = (HereditaryBaseNumber) obj;
			
			// Special optimizations to accelerate comparisons involving 0
			if (n.lowestOrderTerm == null) {
				return this.lowestOrderTerm == null;
			} else if (this.lowestOrderTerm == null) {
				return n.lowestOrderTerm == null;
			}

			// Perform a term-by-term comparison when the radix values match
			// This quickly finds inequalities, but is less
			// efficient than the "slowest comparison" when the objects are
			// equal and the value is cached in both
			if (radix.getValue().equals(n.radix.getValue())) {
				
				Term thisTerm = lowestOrderTerm;
				Term otherTerm = n.lowestOrderTerm;
				while(thisTerm != null && otherTerm != null) {
					// Allow comparison to continue to next term if
					// - The terms both have a next term or both do not
					// - The terms have the same coefficient
					// - The terms have the same exponent (recursive) 
					if (	(thisTerm.next == null && otherTerm.next != null) ||
							(thisTerm.next != null && otherTerm.next == null) ||
							!thisTerm.coefficient.equals(otherTerm.coefficient) ||
							!thisTerm.exponent.equals(otherTerm.exponent)	) {
						return false;
					}
					thisTerm = thisTerm.next;
					otherTerm = otherTerm.next;
				}
				
				return true;
				
			}
						
			// Slowest comparison
			return getValue().equals(n.getValue());
			
		} else if (obj instanceof BigInteger) {
			// Compare to a BigInteger
			BigInteger n = (BigInteger) obj;
						
			// Special optimizations to accelerate comparisons involving 0
			if (n.equals(BigInteger.ZERO)) {
				return this.lowestOrderTerm == null;
			} else if (this.lowestOrderTerm == null) {
				return n.equals(BigInteger.ZERO);
			}
			
			// Special optimization to accelerate comparisons with BigIntegers
			// less than the radix
			if (n.compareTo(radix.getValue()) < 0) {
				if (lowestOrderTerm.exponent.equals(BigInteger.ZERO)
						&& lowestOrderTerm.next == null) {
					return n.equals(lowestOrderTerm.coefficient);
				} else {
					return false;
				}
			}

			// Slowest comparison
			return getValue().equals(n);
			
		} else {
			throw new ClassCastException();
		}
	}

	/**
	 * Returns a string representation of the superbase notation.
	 */
	public String toString() {
		Stack reverseTokens = new Stack();
		Term term = lowestOrderTerm;
		while (term != null) {
			reverseTokens.push(term.toString());
			term = term.next;
			if (term != null) {
				reverseTokens.push(" + ");
			}
		}

		if (reverseTokens.isEmpty()) {
			return "0";
		}
		StringBuffer sb = new StringBuffer();
		while (!reverseTokens.isEmpty()) {
			sb.append((String) reverseTokens.pop());
		}
		return sb.toString();
	}
	
	/**
	 * A simple container class that represents a single term in a linked
	 * list of terms. Each <code>Term</code> stores its coefficient and
	 * exponent. It also provides convenience methods for duplicating subtrees
	 * and generating a string representation. All <code>Term</code>s use
	 * the radix value of the outer <code>HereditaryBaseNumber</code>.
	 * 
	 * @author Matthew Nelson
	 */
	private class Term {
		public BigInteger coefficient;
		public HereditaryBaseNumber exponent;
		public Term next;
		
		public Term(BigInteger c, HereditaryBaseNumber e, Term n) {
			coefficient = c;
			exponent = e;
			next = n;
		}
		
		private Term() {}
		
		/**
		 * Duplicates all subtrees represented by this <code>Term</code>
		 * <i>and all terms chained after it</i>.
		 */
		public Term duplicate() {
			Term clone = new Term();
			
			Term thisTerm = this;
			Term cloneTerm = clone;
			while (thisTerm != null) {
				cloneTerm.coefficient = thisTerm.coefficient;
				cloneTerm.exponent = thisTerm.exponent.duplicate();
				cloneTerm.next = (thisTerm.next == null) ? null : new Term();
				thisTerm = thisTerm.next;
				cloneTerm = cloneTerm.next;
			}
			
			return clone;
		}
		
		/**
		 * Returns a string representation of this term alone. This method
		 * does not read down the linked list like {@link #duplicate()}.
		 */
		public String toString() {
			boolean coefficientIsOne = coefficient.equals(BigInteger.ONE);
			boolean exponentIsZero = exponent.equals(BigInteger.ZERO);
			boolean exponentIsOne = exponent.equals(BigInteger.ONE);
			if (coefficientIsOne && exponentIsZero) {
				return "1";
			} else if (coefficientIsOne && exponentIsOne) {
				return radix.toString();
			} else if (coefficientIsOne) {
				return radix.toString() + "^("
						+ exponent.toString() + ")";
			} else if (exponentIsZero) {
				return coefficient.toString();
			} else if (exponentIsOne) {
				return coefficient.toString() + "*" + radix.toString();
			} else {
				return coefficient + "*" + radix.toString() + "^("
						+ exponent.toString() + ")";
			}
		}
	}

}

/**
 * A very simple class that maintains a <code>BigInteger</code> value and
 * tracks modifications.
 * 
 * @author Matthew Nelson
 */
class SharedValue {
		
	private BigInteger value;
	private int modCount = Integer.MIN_VALUE;
	
	public SharedValue(BigInteger value) {
		this.value = value; 
	}
	
	public void increment() {
		modified();
		value = value.add(BigInteger.ONE);
	}
	
	public void setValue(BigInteger value) {
		modified();
		this.value = value;
	}
	
	public BigInteger getValue() {
		return value;
	}
	
	public int getModCount() {
		return modCount;
	}
	
	private void modified() {
		modCount += 1;
		if (modCount == Integer.MAX_VALUE) {
			modCount = Integer.MIN_VALUE;
		} 
	}
	
	public String toString() {
		return value.toString();
	}
}
